/*
 * gpu_interface.cu
 *
 *  Created on: Oct 8, 2009
 *      Author: Benjamin Block
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#include "gpu_interface.cuh"
#include "gpu_kernel.cuh"
#include <cutil.h>

#include <cstdio>


extern "C"
{

unsigned int gpu_timer;

int gpu_query(int rank)
{
  int devCount;
  cudaGetDeviceCount(&devCount);

//  if (rank < devCount)
//    {
    cudaDeviceProp devProp;
    cudaGetDeviceProperties(&devProp, rank%devCount);

    log_write("Selected %s as GPU", devProp.name);
//    }


  /* here we go */
 // cudaSetDevice(rank%devCount);
  cudaSetDevice(0);
  return devCount;

}

void gpu_alloc_memory(IsingDeviceData* ddata, IsingHostData* hdata)
{
  /* every thread gets one! */
  cudaMalloc((void**) &ddata->randomData,sizeof(int)* NUM_THREADS);

  cudaMalloc((void**) &ddata->metaSpins,sizeof(short)*(N/16));

  /* alloc neighbour arrays */
  cudaMalloc((void**) &ddata->myBorders, sizeof(short)*4*(n/4));
  cudaMalloc((void**) &ddata->theirBorders, sizeof(short)*4*(n/4));

  /* allocate output values array */
  cudaMalloc((void**) &ddata->out, sizeof(int)*NUM_BLOCKS);

  /* alloc host memory that is not transferred frequently */
  hdata->metaSpins = new unsigned short[N/16];
  hdata->randomData = new int[NUM_THREADS];

  /* allocate pinned host memory - or normal if wished */
#ifdef PINNED
  cudaHostAlloc((void**) &hdata->out,sizeof(int)*NUM_BLOCKS, cudaHostAllocPortable);
  cudaHostAlloc((void**) &hdata->theirBorders, sizeof(short)*4*(n/4), cudaHostAllocPortable);
  cudaHostAlloc((void**) &hdata->myBorders, sizeof(short)*4*(n/4), cudaHostAllocPortable);
#else
  hdata->out = new int[NUM_BLOCKS];
  hdata->theirBorders = new unsigned short[4*(n/4)];
  hdata->myBorders = new unsigned short[4*(n/4)];
#endif

  memset(hdata->metaSpins, 0x00, sizeof(short) * N /16);
  memset(hdata->theirBorders, 0x00, sizeof(short) * n /4 * 4);
}

void gpu_free_memory(IsingDeviceData* ddata, IsingHostData* hdata)
{
  cudaFree(ddata->randomData);
  cudaFree(ddata->metaSpins);
  cudaFree(ddata->out);

  // free neighbour arrays
  cudaFree(ddata->myBorders);
  cudaFree(ddata->theirBorders);

  delete[] hdata->metaSpins;
  delete[] hdata->randomData;

#ifdef PINNED
  cudaFreeHost(hdata->out);
  cudaFreeHost(hdata->theirBorders);
  cudaFreeHost(hdata->myBorders);
#else
  delete[] hdata->out;
  delete[] hdata->theirBorders;
  delete[] hdata->myBorders;
#endif
}

void gpu_data_deposit(IsingDeviceData* ddata, IsingHostData* hdata)
{
  cudaMemcpy(ddata->randomData,hdata->randomData,sizeof(int)*NUM_THREADS,cudaMemcpyHostToDevice);
  cudaMemcpy(ddata->metaSpins, hdata->metaSpins, sizeof(short)*(N/16),cudaMemcpyHostToDevice);
}
void gpu_out_retrieve(IsingDeviceData* ddata, IsingHostData* hdata)
{
  cudaMemcpy(hdata->out,ddata->out,sizeof(int)*FIELD_DIM * FIELD_DIM,cudaMemcpyDeviceToHost);
}

void gpu_spins_retrieve(IsingDeviceData* ddata, IsingHostData* hdata)
{
  cudaMemcpy(hdata->metaSpins,ddata->metaSpins,sizeof(short)*(N/16),cudaMemcpyDeviceToHost);
}

void gpu_borders_retrieve(IsingDeviceData* ddata, IsingHostData *hdata)
{
  cudaMemcpy(hdata->myBorders,ddata->myBorders,sizeof(short)*4*(n/4),cudaMemcpyDeviceToHost);
}

void gpu_borders_deposit(IsingDeviceData* ddata, IsingHostData *hdata)
{
  cudaMemcpy(ddata->theirBorders, hdata->theirBorders, sizeof(short)*4*(n/4),cudaMemcpyHostToDevice);
}

void gpu_timer_start()
{
  cudaThreadSynchronize();
  cutCreateTimer(&gpu_timer);
  cutStartTimer(gpu_timer);
}

float gpu_timer_stop()
{
  cudaThreadSynchronize();
  cutStopTimer(gpu_timer);
  float result = cutGetTimerValue(gpu_timer);
  cutDeleteTimer(gpu_timer);
  return result;

}

void gpu_kernel_call(IsingDeviceData* ddata, float t, bool flag)
{
  dim3 grid(FIELD_DIM, FIELD_DIM);
  dim3 threads(BLOCK_DIM, BLOCK_DIM);
  gpu_kernel_checkerboard<<<grid, threads>>>(*ddata, t, flag);
  cudaThreadSynchronize();

#ifdef MULTI_GPU
  gpu_kernel_prepare_borders<<<8, BLOCK_DIM * FIELD_DIM>>>(*ddata);
#endif
}
}
