/*
 * gpu_interface.cuh
 *
 * This file defines everything needed to interface the GPU
 *
 *  Created on: Oct 8, 2009
 *      Author: Benjamin Block
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#ifndef GPU_INTERFACE_CUH_
#define GPU_INTERFACE_CUH_

#include "core.h"

/* useful for implementation */
#define NUM_THREADS (BLOCK_DIM * BLOCK_DIM * FIELD_DIM * FIELD_DIM)
#define NUM_BLOCKS (FIELD_DIM * FIELD_DIM)

/* This datastructure holds all data handed to the GPU on a kernel call.
 * This way the GPU knows where to find the data in the GPU memory
 */
struct IsingDeviceData
{
  unsigned short *metaSpins; /* (4x4 spins) in each short */
  int *randomData; /* random seeds for each thread */
  int *out; /* out array, stores summed magetization */

  /* the borders have to be updated after exchange */
  unsigned short *theirBorders;
  unsigned short *myBorders;

  /* precalculated exp factors */
  float exp_dH_4;
  float exp_dH_8;
};

/* forward declaration */
struct IsingHostData;

/* function declarations */
extern "C"
{

/* needed for the timer */
extern unsigned int gpu_timer;

/* deposits the prepared data on the GPU */
void  gpu_data_deposit(IsingDeviceData* ddata, IsingHostData* hdata);
/* retrieve output values (magnetization) from GPU */
void  gpu_out_retrieve(IsingDeviceData* ddata, IsingHostData* hdata);
/* retrieve spin field from GPU */
void  gpu_spins_retrieve(IsingDeviceData* ddata, IsingHostData* hdata);
/* store exchanged borders to GPU memory */
void  gpu_borders_deposit(IsingDeviceData *ddata, IsingHostData* hdata);
/* retrieve borders from GPU for exchange */
void  gpu_borders_retrieve(IsingDeviceData *ddata, IsingHostData* hdata);
/* allocate necessary memory on GPU */
void  gpu_alloc_memory(IsingDeviceData* ddata, IsingHostData *hdata);
/* free all memory on GPU */
void  gpu_free_memory(IsingDeviceData* ddata, IsingHostData *hdata);
/* call the update kernel. flag = false executes first step, flag = true executes second step */
void  gpu_kernel_call(IsingDeviceData* ddata, float t, bool flag);

/* start the timer from the CUDA SDK */
void  gpu_timer_start();
/* stop the timer from the CUDA SDK */
float gpu_timer_stop();

/* query the GPU */
int   gpu_query(int rank);
}

#endif /* GPU_INTERFACE_CUH_ */
