/*
 * gpu_kernel.cu
 *
 *  Created on: Oct 8, 2009
 *      Author: Benjamin Block
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#include "gpu_interface.cuh"
#include "gpu_kernel.cuh"
#include <cstdio>

/* Helper macro to access a two dimensional (meta) field
 */
#define _F(_x,_y) ((_x) + (n/4) * (_y))

#define _S(meta,_x,_y) (((meta & ( 1 << ((_y) * 4 + (_x)))) != 0) * 2 -1)

#define _SF(_x,_y) ((_x) + (6 * (_y)))

#define SPIN_PROC(_x,_y) \
rand = RANDOM_A * rand + RANDOM_B; \
      dH = 2 * field[_SF(_x,_y)] * (field[_SF(_x-1,_y)] + \
          field[_SF(_x+1,_y)] + \
          field[_SF(_x,_y-1)] + \
          field[_SF(_x,_y+1)]); \
      /* set the update flag */                \
      if ((dH <= 0)                \
          ||((dH == 4) && (fabs(rand*4.656612e-10f)<ddata.exp_dH_4))  \
          ||((dH == 8) && (fabs(rand*4.656612e-10f)<ddata.exp_dH_8))  \
              )                \
              { \
               update_pattern |= (1 << ((_y)-1) * 4 + ((_x)-1));  \
               field[_SF(_x,_y)] = -field[_SF(_x,_y)]; \
               }



/* The main work is done here
 * This updates the spin field in two steps, as described
 * flag = 0 executes the first step
 * flag = 1 executes the second step
 */


__global__ void gpu_kernel_checkerboard(IsingDeviceData ddata, float t, bool flag) {

  // this is the temporary field
  __shared__ char shared_field[6*6*BLOCK_DIM * BLOCK_DIM];
  char *field = &shared_field[6*6* (threadIdx.y * BLOCK_DIM + threadIdx.x)];

  // the update pattern
  unsigned short update_pattern = 0x00;

  // coordinate system
  int x = threadIdx.x + blockIdx.x * BLOCK_DIM; // MAX: FIELD_DIM * BLOCK_DIM
  int y = threadIdx.y + blockIdx.y * BLOCK_DIM; // MAX FIELD_DIM * BLOCK_DIM

//  ONLY_ROOT(3,1) printf("x = %i, y = %i", x,y);

  // running indices
  int i,j;

  __shared__ int r[BLOCK_DIM*BLOCK_DIM];
  int rand = ddata.randomData[x + BLOCK_DIM*FIELD_DIM *y];

 // ONLY_ROOT(3,1) printf("rand10 = %i\n", rand10);

  // retrieve the only 5 needed metaSpins for the 4x4 spin field
  unsigned short me = ddata.metaSpins[_F(x,y)];

//  __syncthreads();

//  ONLY_ROOT(3,1) printf("spins me %hu\n",ddata.metaSpins[_F(x,y)]);

  unsigned short above;
  unsigned short below;
  unsigned short left;
  unsigned short right;

  /* catch the upper border */
  if (y == 0)
    {
#ifdef MULTI_GPU
      above = ddata.theirBorders[N_TOP*(n/4) + x];
#else
      above = ddata.metaSpins[_F(x,(n/4)-1)];
#endif
    }
  else
      above = ddata.metaSpins[_F(x,y-1)];

  /* catch the lower border */
  if (y == n/4 - 1)
    {
#ifdef MULTI_GPU
      below = ddata.theirBorders[N_BOTTOM*(n/4) + x];
#else
      below = ddata.metaSpins[_F(x,0)];
 //     ONLY_ROOT(3,1) printf("Edge warp");
#endif
    }
  else
      below = ddata.metaSpins[_F(x,y+1)];


  if (x == 0)
    {
#ifdef MULTI_GPU
      left = ddata.theirBorders[N_LEFT*(n/4) + y];
#else
      left = ddata.metaSpins[_F((n/4)-1,y)];
#endif
    }
  else
      left = ddata.metaSpins[_F(x-1,y)];


  /* catch the lower border */
  if (x == n/4 - 1)
    {
#ifdef MULTI_GPU
      right = ddata.theirBorders[N_RIGHT*(n/4) + y];
#else
      right = ddata.metaSpins[_F(0,y)];
#endif
    }
  else
      right = ddata.metaSpins[_F(x+1,y)];

  /* fill the field with the dense coded spins */

  /* top */
  for (i = 1; i < 5; i++)
    {
    field[_SF(0,i)] = _S(left,3,i-1);
    field[_SF(5,i)] = _S(right,0,i-1);
    field[_SF(i,5)] = _S(below,i-1,0);
    field[_SF(i,0)] = _S(above,i-1,3);
    }

  /* extract my own spins */
  for ( j = 1; j < 5; j++)
    for (i = 1; i < 5; i++)
      {
      field[_SF(i,j)] = _S(me, i-1,j-1);
      }

  /* process the "inner spins" */
  int dH;

  if (flag)
    {
      /* update white sites */
      SPIN_PROC(1,1)    SPIN_PROC(2,2)
      SPIN_PROC(3,1)    SPIN_PROC(4,2)
      SPIN_PROC(1,3)    SPIN_PROC(2,4)
      SPIN_PROC(3,3)    SPIN_PROC(4,4)
    }
  else
    {
      SPIN_PROC(2,1)    SPIN_PROC(1,2)
      SPIN_PROC(4,1)    SPIN_PROC(3,2)
      SPIN_PROC(2,3)    SPIN_PROC(1,4)
      SPIN_PROC(4,3)    SPIN_PROC(3,4)

    dH = 0;

    /* add up magnetisation */
    for ( i = 1; i < 5; i++)
      for (j = 1; j < 5; j++)
        dH += field[_SF(i,j)];


    r[(threadIdx.y * BLOCK_DIM + threadIdx.x)] = dH;

    __syncthreads();

    for(int dx=1;dx<BLOCK_DIM*BLOCK_DIM;dx*=2) {
        if((threadIdx.y * BLOCK_DIM + threadIdx.x)%(2*dx)==0) {
           r[(threadIdx.y * BLOCK_DIM + threadIdx.x)]+=r[(threadIdx.y * BLOCK_DIM + threadIdx.x)+dx];
         }
         __syncthreads();
    }

    /* save into out array */
    if ((threadIdx.x == 0 ) && ( threadIdx.y == 0))
     {
       ddata.out[blockIdx.y * FIELD_DIM + blockIdx.x] = r[0];
     }
    }

  ddata.randomData[x + BLOCK_DIM*FIELD_DIM*y] = rand;

  /* update spin field */
  ddata.metaSpins[_F(x,y)] ^= update_pattern;
}

__global__ void gpu_kernel_prepare_borders(IsingDeviceData ddata)
{
  int i = threadIdx.x+BLOCK_DIM * FIELD_DIM *blockIdx.x;

  int b = i / (n/4);
  i = i % (n/4);

  switch (b)
  {
  case 0:
    ddata.myBorders[N_TOP*(n/4)+i] = ddata.metaSpins[i];
    break;
  case 1:
    ddata.myBorders[N_BOTTOM*(n/4)+i] = ddata.metaSpins[(N/16) - (n/4) + i];
    break;
  case 2:
    ddata.myBorders[N_LEFT*(n/4)+i] = ddata.metaSpins[i*(n/4)];
    break;
  case 3:
    ddata.myBorders[N_RIGHT*(n/4)+i] = ddata.metaSpins[(i+1)*(n/4) - 1];
    break;
  }
}
