/*
 * mpi_communication.cpp
 *
 * In this file, MPI communication is implemented.
 *
 *  Created on: Oct 9, 2009
 *      Author: Benjamin Block
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#include <mpi.h>

#include "mpi_communication.h"

int rank = 0;
int nthreads = 1;

#ifdef MULTI_GPU
#include <cstdlib>


void mpi_init(int argc, char **argv)
{
  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nthreads);

  MPI_Barrier(MPI_COMM_WORLD);
}

/* Exchanging the borders of the spin field via MPI
 *
 */
void mpi_exchange_borders(IsingHostData *hdata, int myrank, int *myneighbours)
{
  MPI_Request request[8];

  /* send my borders */
  MPI_Isend(&hdata->myBorders[N_LEFT*(n/4)], n/4, MPI_SHORT, myneighbours[N_LEFT], N_LEFT, MPI_COMM_WORLD, &request[0]);
  MPI_Isend(&hdata->myBorders[N_RIGHT*(n/4)], n/4, MPI_SHORT, myneighbours[N_RIGHT], N_RIGHT, MPI_COMM_WORLD, &request[1]);
  MPI_Isend(&hdata->myBorders[N_TOP*(n/4)], n/4, MPI_SHORT, myneighbours[N_TOP], N_TOP, MPI_COMM_WORLD, &request[2]);
  MPI_Isend(&hdata->myBorders[N_BOTTOM*(n/4)], n/4, MPI_SHORT, myneighbours[N_BOTTOM], N_BOTTOM, MPI_COMM_WORLD, &request[3]);

  /* receive their borders */

  /* note: the flags have to be reversed, since the left neighbour sent his right edge etc. */
  MPI_Irecv(&hdata->theirBorders[N_LEFT*(n/4)], n/4, MPI_SHORT, myneighbours[N_LEFT], N_RIGHT, MPI_COMM_WORLD, &request[4]);
  MPI_Irecv(&hdata->theirBorders[N_RIGHT*(n/4)], n/4, MPI_SHORT, myneighbours[N_RIGHT], N_LEFT, MPI_COMM_WORLD, &request[5]);
  MPI_Irecv(&hdata->theirBorders[N_TOP*(n/4)], n/4, MPI_SHORT, myneighbours[N_TOP], N_BOTTOM, MPI_COMM_WORLD, &request[6]);
  MPI_Irecv(&hdata->theirBorders[N_BOTTOM*(n/4)], n/4, MPI_SHORT, myneighbours[N_BOTTOM], N_TOP, MPI_COMM_WORLD, &request[7]);

  /* wait for all requests to complete */
  MPI_Waitall(8, request, MPI_STATUSES_IGNORE);
}

void mpi_reduce_result(double *result, int numEntries)
{
  double *receive = new double[numEntries];

  MPI_Reduce(result, receive, numEntries, MPI_DOUBLE, MPI_SUM, 0,  MPI_COMM_WORLD);

  /* divide by number of runs */
  int i;
  for (i = 0; i < numEntries; i++)
    {
    result[i] = receive[i] / nthreads;
    }

  delete[] receive;
}

#endif
